import random
import datetime as dt

from pycalver import version
from pycalver import patterns


def test_bump_beta():
    cur_version = "v201712.0001-beta"
    assert cur_version < version.incr(cur_version)
    assert version.incr(cur_version).endswith("-beta")
    assert version.incr(cur_version, release="alpha").endswith("-alpha")
    assert version.incr(cur_version, release="final").endswith("0002")


def test_bump_final():
    cur_version = "v201712.0001"
    assert cur_version < version.incr(cur_version)
    assert version.incr(cur_version).endswith(".0002")
    assert version.incr(cur_version, release="alpha").endswith("-alpha")

    assert version.incr(cur_version, release="final").endswith(".0002")

    pre_version = cur_version + "-beta"
    assert version.incr(pre_version, release="final").endswith(".0002")


def test_bump_future():
    """Test that versions don't go back in time."""
    future_date   = dt.datetime.today() + dt.timedelta(days=300)
    future_calver = future_date.strftime("v%Y%m")
    cur_version   = future_calver + ".0001"
    new_version   = version.incr(cur_version)
    assert cur_version < new_version


def test_bump_random(monkeypatch):
    cur_date    = dt.date(2016, 1, 1) + dt.timedelta(days=random.randint(1, 2000))
    cur_version = cur_date.strftime("v%Y%m") + ".0001-dev"

    monkeypatch.setattr(version, 'TODAY', cur_date)

    for i in range(1000):
        cur_date += dt.timedelta(days=int((1 + random.random()) ** 10))
        new_version = version.incr(
            cur_version, release=random.choice([None, "alpha", "beta", "rc", "final", "post"])
        )
        assert cur_version < new_version
        cur_version = new_version


def test_parse_version_info():
    version_str = "v201712.0001-alpha"
    version_nfo = version.parse_version_info(version_str)

    # assert version_nfo.pep440_version == "201712.1a0"
    # assert version_nfo.version        == "v201712.0001-alpha"
    assert version_nfo.year  == 2017
    assert version_nfo.month == 12
    assert version_nfo.bid   == "0001"
    assert version_nfo.tag   == "alpha"

    version_str = "v201712.0001"
    version_nfo = version.parse_version_info(version_str)

    # assert version_nfo.pep440_version == "201712.1"
    # assert version_nfo.version        == "v201712.0001"
    assert version_nfo.year  == 2017
    assert version_nfo.month == 12
    assert version_nfo.bid   == "0001"
    assert version_nfo.tag   == "final"


def test_readme_pycalver1():
    version_str  = "v201712.0001-alpha"
    version_info = patterns.PYCALVER_RE.match(version_str).groupdict()

    assert version_info == {
        'pycalver'   : "v201712.0001-alpha",
        'vYYYYMM'    : "v201712",
        'year'       : "2017",
        'month'      : "12",
        'build'      : ".0001",
        'build_no'   : "0001",
        'release'    : "-alpha",
        'release_tag': "alpha",
    }


def test_readme_pycalver2():
    version_str  = "v201712.0033"
    version_info = patterns.PYCALVER_RE.match(version_str).groupdict()

    assert version_info == {
        'pycalver'   : "v201712.0033",
        'vYYYYMM'    : "v201712",
        'year'       : "2017",
        'month'      : "12",
        'build'      : ".0033",
        'build_no'   : "0033",
        'release'    : None,
        'release_tag': None,
    }


def test_parse_error_empty():
    try:
        version.parse_version_info("")
        assert False
    except ValueError as err:
        pass


def test_parse_error_noprefix():
    try:
        version.parse_version_info("201809.0002")
        assert False
    except ValueError as err:
        pass


def test_parse_error_nopadding():
    try:
        version.parse_version_info("v201809.2b0")
        assert False
    except ValueError as err:
        pass


def test_part_field_mapping():
    a_names = set(version.PATTERN_PART_FIELDS.keys())
    b_names = set(patterns.PART_PATTERNS.keys())
    c_names = set(patterns.COMPOSITE_PART_PATTERNS.keys())

    extra_names = a_names - b_names
    assert not any(extra_names)
    missing_names = b_names - a_names
    assert missing_names == c_names

    a_fields = set(version.PATTERN_PART_FIELDS.values())
    b_fields = set(version.VersionInfo._fields)

    assert a_fields == b_fields



# This file is part of the pycalver project
# https://gitlab.com/mbarkhau/pycalver
#
# Copyright (c) 2019 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT
"""Functions related to version string manipulation."""

import logging
import pkg_resources
import typing as typ
import datetime as dt

from . import lex_id
from . import patterns

log = logging.getLogger("pycalver.version")


# The test suite may replace this.
TODAY = dt.datetime.utcnow().date()


PATTERN_PART_FIELDS = {
    'year'      : 'year',
    'month'     : 'month',
    'pep440_tag': 'tag',
    'tag'       : 'tag',
    'yy'        : 'year',
    'yyyy'      : 'year',
    'quarter'   : 'quarter',
    'iso_week'  : 'iso_week',
    'us_week'   : 'us_week',
    'dom'       : 'dom',
    'doy'       : 'doy',
    'MAJOR'     : 'major',
    'MINOR'     : 'minor',
    'MM'        : 'minor',
    'MMM'       : 'minor',
    'MMMM'      : 'minor',
    'MMMMM'     : 'minor',
    'PP'        : 'patch',
    'PPP'       : 'patch',
    'PPPP'      : 'patch',
    'PPPPP'     : 'patch',
    'PATCH'     : 'patch',
    'build_no'  : 'bid',
    'bid'       : 'bid',
    'BID'       : 'bid',
    'BB'        : 'bid',
    'BBB'       : 'bid',
    'BBBB'      : 'bid',
    'BBBBB'     : 'bid',
    'BBBBBB'    : 'bid',
    'BBBBBBB'   : 'bid',
}


class CalendarInfo(typ.NamedTuple):
    """Container for calendar components of version strings."""

    year    : int
    quarter : int
    month   : int
    dom     : int
    doy     : int
    iso_week: int
    us_week : int


def _date_from_doy(year: int, doy: int) -> dt.date:
    """Parse date from year and day of year (1 indexed).

    >>> cases = [
    ...     (2016, 1), (2016, 31), (2016, 31 + 1), (2016, 31 + 29), (2016, 31 + 30),
    ...     (2017, 1), (2017, 31), (2017, 31 + 1), (2017, 31 + 28), (2017, 31 + 29),
    ... ]
    >>> dates = [_date_from_doy(year, month) for year, month in cases]
    >>> assert [(d.month, d.day) for d in dates] == [
    ...     (1, 1), (1, 31), (2, 1), (2, 29), (3, 1),
    ...     (1, 1), (1, 31), (2, 1), (2, 28), (3, 1),
    ... ]
    """
    return dt.date(year, 1, 1) + dt.timedelta(days=doy - 1)


def _quarter_from_month(month: int) -> int:
    """Calculate quarter (1 indexed) from month (1 indexed).

    >>> [_quarter_from_month(month) for month in range(1, 13)]
    [1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4]
    """
    return ((month - 1) // 3) + 1


def cal_info(date: dt.date = None) -> CalendarInfo:
    """Generate calendar components for current date.

    >>> from datetime import date

    >>> c = cal_info(date(2019, 1, 5))
    >>> (c.year, c.quarter, c.month, c.dom, c.doy, c.iso_week, c.us_week)
    (2019, 1, 1, 5, 5, 0, 0)

    >>> c = cal_info(date(2019, 1, 6))
    >>> (c.year, c.quarter, c.month, c.dom, c.doy, c.iso_week, c.us_week)
    (2019, 1, 1, 6, 6, 0, 1)

    >>> c = cal_info(date(2019, 1, 7))
    >>> (c.year, c.quarter, c.month, c.dom, c.doy, c.iso_week, c.us_week)
    (2019, 1, 1, 7, 7, 1, 1)

    >>> c = cal_info(date(2019, 4, 7))
    >>> (c.year, c.quarter, c.month, c.dom, c.doy, c.iso_week, c.us_week)
    (2019, 2, 4, 7, 97, 13, 14)
    """
    if date is None:
        date = TODAY

    kw = {
        'year'    : date.year,
        'quarter' : _quarter_from_month(date.month),
        'month'   : date.month,
        'dom'     : date.day,
        'doy'     : int(date.strftime("%j"), base=10),
        'iso_week': int(date.strftime("%W"), base=10),
        'us_week' : int(date.strftime("%U"), base=10),
    }

    return CalendarInfo(**kw)


class VersionInfo(typ.NamedTuple):
    """Container for parsed version string."""

    year    : typ.Optional[int]
    quarter : typ.Optional[int]
    month   : typ.Optional[int]
    dom     : typ.Optional[int]
    doy     : typ.Optional[int]
    iso_week: typ.Optional[int]
    us_week : typ.Optional[int]
    major   : int
    minor   : int
    patch   : int
    bid     : str
    tag     : str


def _is_calver(nfo: typ.Union[CalendarInfo, VersionInfo]) -> bool:
    """Check pattern for any calendar based parts.

    >>> _is_calver(cal_info())
    True

    >>> vnfo = _parse_version_info({'year': "2018", 'month': "11", 'bid': "0018"})
    >>> _is_calver(vnfo)
    True

    >>> vnfo = _parse_version_info({'MAJOR': "1", 'MINOR': "023", 'PATCH': "45"})
    >>> _is_calver(vnfo)
    False
    """
    for field in CalendarInfo._fields:
        if isinstance(getattr(nfo, field, None), int):
            return True

    return False


TAG_ALIASES: typ.Dict[str, str] = {'a': "alpha", 'b': "beta", 'pre': "rc"}


PEP440_TAGS: typ.Dict[str, str] = {
    'alpha': "a",
    'beta' : "b",
    'final': "",
    'rc'   : "rc",
    'dev'  : "dev",
    'post' : "post",
}


VersionInfoKW = typ.Dict[str, typ.Union[str, int, None]]


def _parse_pattern_groups(pattern_groups: typ.Dict[str, str]) -> typ.Dict[str, str]:
    for part_name in pattern_groups.keys():
        is_valid_part_name = (
            part_name in patterns.COMPOSITE_PART_PATTERNS or part_name in PATTERN_PART_FIELDS
        )
        if not is_valid_part_name:
            err_msg = f"Invalid part '{part_name}'"
            raise ValueError(err_msg)

    items = [
        (field_name, pattern_groups[part_name])
        for part_name, field_name in PATTERN_PART_FIELDS.items()
        if part_name in pattern_groups.keys()
    ]
    all_fields       = [field_name for field_name, _ in items]
    unique_fields    = set(all_fields)
    duplicate_fields = [f for f in unique_fields if all_fields.count(f) > 1]

    if any(duplicate_fields):
        err_msg = f"Multiple parts for same field {duplicate_fields}."
        raise ValueError(err_msg)

    return dict(items)


def _parse_version_info(pattern_groups: typ.Dict[str, str]) -> VersionInfo:
    """Parse normalized VersionInfo from groups of a matched pattern.

    >>> vnfo = _parse_version_info({'year': "2018", 'month': "11", 'bid': "0099"})
    >>> (vnfo.year, vnfo.month, vnfo.quarter, vnfo.bid, vnfo.tag)
    (2018, 11, 4, '0099', 'final')

    >>> vnfo = _parse_version_info({'year': "2018", 'doy': "11", 'bid': "099", 'tag': "b"})
    >>> (vnfo.year, vnfo.month, vnfo.dom, vnfo.bid, vnfo.tag)
    (2018, 1, 11, '099', 'beta')

    >>> vnfo = _parse_version_info({'MAJOR': "1", 'MINOR': "23", 'PATCH': "45"})
    >>> (vnfo.major, vnfo.minor, vnfo.patch)
    (1, 23, 45)

    >>> vnfo = _parse_version_info({'MAJOR': "1", 'MMM': "023", 'PPPP': "0045"})
    >>> (vnfo.major, vnfo.minor, vnfo.patch)
    (1, 23, 45)
    """
    kw = _parse_pattern_groups(pattern_groups)

    tag = kw.get('tag')
    if tag is None:
        tag = "final"
    tag = TAG_ALIASES.get(tag, tag)
    assert tag is not None

    bid = kw['bid'] if 'bid' in kw else "0001"

    year = int(kw['year']) if 'year' in kw else None
    doy  = int(kw['doy' ]) if 'doy' in kw else None

    month: typ.Optional[int]
    dom  : typ.Optional[int]

    if year and doy:
        date  = _date_from_doy(year, doy)
        month = date.month
        dom   = date.day
    else:
        month = int(kw['month']) if 'month' in kw else None
        dom   = int(kw['dom'  ]) if 'dom' in kw else None

    iso_week: typ.Optional[int]
    us_week : typ.Optional[int]

    if year and month and dom:
        date     = dt.date(year, month, dom)
        doy      = int(date.strftime("%j"), base=10)
        iso_week = int(date.strftime("%W"), base=10)
        us_week  = int(date.strftime("%U"), base=10)
    else:
        iso_week = None
        us_week  = None

    quarter = int(kw['quarter']) if 'quarter' in kw else None
    if quarter is None and month:
        quarter = _quarter_from_month(month)

    major = int(kw['major']) if 'major' in kw else 0
    minor = int(kw['minor']) if 'minor' in kw else 0
    patch = int(kw['patch']) if 'patch' in kw else 0

    return VersionInfo(
        year=year,
        quarter=quarter,
        month=month,
        dom=dom,
        doy=doy,
        iso_week=iso_week,
        us_week=us_week,
        major=major,
        minor=minor,
        patch=patch,
        bid=bid,
        tag=tag,
    )


def parse_version_info(version_str: str, pattern: str = "{pycalver}") -> VersionInfo:
    """Parse normalized VersionInfo.

    >>> vnfo = parse_version_info("v201712.0033-beta", pattern="{pycalver}")
    >>> assert vnfo == _parse_version_info({'year': 2017, 'month': 12, 'bid': "0033", 'tag': "beta"})

    >>> vnfo = parse_version_info("1.23.456", pattern="{semver}")
    >>> assert vnfo == _parse_version_info({'MAJOR': "1", 'MINOR': "23", 'PATCH': "456"})
    """
    regex = patterns.compile_pattern(pattern)
    match = regex.match(version_str)
    if match is None:
        err_msg = (
            f"Invalid version string '{version_str}' for pattern '{pattern}'/'{regex.pattern}'"
        )
        raise ValueError(err_msg)

    return _parse_version_info(match.groupdict())


def is_valid(version_str: str, pattern: str = "{pycalver}") -> bool:
    """Check if a version matches a pattern.

    >>> is_valid("v201712.0033-beta", pattern="{pycalver}")
    True
    >>> is_valid("v201712.0033-beta", pattern="{semver}")
    False
    >>> is_valid("1.2.3", pattern="{semver}")
    True
    >>> is_valid("v201712.0033-beta", pattern="{semver}")
    False
    """
    try:
        parse_version_info(version_str, pattern)
        return True
    except ValueError:
        return False


ID_FIELDS_BY_PART = {
    'MAJOR'  : 'major',
    'MINOR'  : 'minor',
    'MM'     : 'minor',
    'MMM'    : 'minor',
    'MMMM'   : 'minor',
    'MMMMM'  : 'minor',
    'MMMMMM' : 'minor',
    'MMMMMMM': 'minor',
    'PATCH'  : 'patch',
    'PP'     : 'patch',
    'PPP'    : 'patch',
    'PPPP'   : 'patch',
    'PPPPP'  : 'patch',
    'PPPPPP' : 'patch',
    'PPPPPPP': 'patch',
    'BID'    : 'bid',
    'BB'     : 'bid',
    'BBB'    : 'bid',
    'BBBB'   : 'bid',
    'BBBBB'  : 'bid',
    'BBBBBB' : 'bid',
    'BBBBBBB': 'bid',
}


def format_version(ver_nfo: VersionInfo, pattern: str) -> str:
    """Generate version string.

    >>> import datetime as dt
    >>> ver_nfo = parse_version_info("v201712.0033-beta", pattern="{pycalver}")
    >>> ver_nfo_a = ver_nfo._replace(**cal_info(date=dt.date(2017, 1, 1))._asdict())
    >>> ver_nfo_b = ver_nfo._replace(**cal_info(date=dt.date(2017, 12, 31))._asdict())
    >>> ver_nfo_c = ver_nfo_b._replace(major=1, minor=2, patch=34, tag='final')

    >>> format_version(ver_nfo_a, pattern="v{yy}.{BID}{release}")
    'v17.33-beta'
    >>> format_version(ver_nfo_a, pattern="{pep440_version}")
    '201701.33b0'

    >>> format_version(ver_nfo_a, pattern="{pycalver}")
    'v201701.0033-beta'
    >>> format_version(ver_nfo_b, pattern="{pycalver}")
    'v201712.0033-beta'

    >>> format_version(ver_nfo_a, pattern="v{year}w{iso_week}.{BID}{release}")
    'v2017w00.33-beta'
    >>> format_version(ver_nfo_b, pattern="v{year}w{iso_week}.{BID}{release}")
    'v2017w52.33-beta'

    >>> format_version(ver_nfo_a, pattern="v{year}d{doy}.{bid}{release}")
    'v2017d001.0033-beta'
    >>> format_version(ver_nfo_b, pattern="v{year}d{doy}.{bid}{release}")
    'v2017d365.0033-beta'

    >>> format_version(ver_nfo_c, pattern="v{year}w{iso_week}.{BID}-{tag}")
    'v2017w52.33-final'
    >>> format_version(ver_nfo_c, pattern="v{year}w{iso_week}.{BID}{release}")
    'v2017w52.33'

    >>> format_version(ver_nfo_c, pattern="v{MAJOR}.{MINOR}.{PATCH}")
    'v1.2.34'
    >>> format_version(ver_nfo_c, pattern="v{MAJOR}.{MM}.{PPP}")
    'v1.02.034'
    """
    full_pattern = pattern
    for part_name, full_part_format in patterns.FULL_PART_FORMATS.items():
        full_pattern = full_pattern.replace("{" + part_name + "}", full_part_format)

    kw = ver_nfo._asdict()
    if kw['tag'] == 'final':
        kw['release'   ] = ""
        kw['pep440_tag'] = ""
    else:
        kw['release'   ] = "-" + kw['tag']
        kw['pep440_tag'] = PEP440_TAGS[kw['tag']] + "0"

    kw['release_tag'] = kw['tag']

    kw['yy'  ] = str(kw['year'])[-2:]
    kw['yyyy'] = kw['year']
    kw['BID' ] = int(kw['bid'], 10)

    for part_name, field in ID_FIELDS_BY_PART.items():
        val = kw[field]
        if part_name.lower() == field.lower():
            if isinstance(val, str):
                kw[part_name] = int(val, base=10)
            else:
                kw[part_name] = val
        else:
            assert len(set(part_name)) == 1
            padded_len = len(part_name)
            kw[part_name] = str(val).zfill(padded_len)

    return full_pattern.format(**kw)


def incr(
    old_version: str,
    pattern    : str = "{pycalver}",
    *,
    release: str  = None,
    major  : bool = False,
    minor  : bool = False,
    patch  : bool = False,
) -> typ.Optional[str]:
    """Increment version string.

    'old_version' is assumed to be a string that matches 'pattern'
    """
    try:
        old_ver_nfo = parse_version_info(old_version, pattern)
    except ValueError as ex:
        log.error(str(ex))
        return None

    cur_ver_nfo = old_ver_nfo

    cur_cal_nfo = cal_info()

    old_date = (old_ver_nfo.year or 0, old_ver_nfo.month or 0, old_ver_nfo.dom or 0)
    cur_date = (cur_cal_nfo.year     , cur_cal_nfo.month     , cur_cal_nfo.dom)

    if old_date <= cur_date:
        cur_ver_nfo = cur_ver_nfo._replace(**cur_cal_nfo._asdict())
    else:
        log.warning(f"Version appears to be from the future '{old_version}'")

    cur_ver_nfo = cur_ver_nfo._replace(bid=lex_id.next_id(cur_ver_nfo.bid))

    if major:
        cur_ver_nfo = cur_ver_nfo._replace(major=cur_ver_nfo.major + 1, minor=0, patch=0)
    if minor:
        cur_ver_nfo = cur_ver_nfo._replace(minor=cur_ver_nfo.minor + 1, patch=0)
    if patch:
        cur_ver_nfo = cur_ver_nfo._replace(patch=cur_ver_nfo.patch + 1)

    if release:
        cur_ver_nfo = cur_ver_nfo._replace(tag=release)

    new_version = format_version(cur_ver_nfo, pattern)
    if new_version == old_version:
        log.error("Invalid arguments or pattern, version did not change.")
        return None
    else:
        return new_version


def to_pep440(version: str) -> str:
    """Derive pep440 compliant version string from PyCalVer version string.

    >>> to_pep440("v201811.0007-beta")
    '201811.7b0'
    """
    return str(pkg_resources.parse_version(version))



import re
import pytest

from pycalver import patterns


def _part_re_by_name(name):
    return re.compile(patterns.PART_PATTERNS[name])


@pytest.mark.parametrize("part_name", patterns.PART_PATTERNS.keys())
def test_part_compilation(part_name):
    assert _part_re_by_name(part_name)


PATTERN_PART_CASES = [
    ("pep440_pycalver", "201712.31"          , "201712.31"),
    ("pep440_pycalver", "v201712.0032"       , None),
    ("pep440_pycalver", "201712.0033-alpha"  , None),
    ("pycalver"       , "v201712.0034"       , "v201712.0034"),
    ("pycalver"       , "v201712.0035-alpha" , "v201712.0035-alpha"),
    ("pycalver"       , "v201712.0036-alpha0", "v201712.0036-alpha"),
    ("pycalver"       , "v201712.0037-pre"   , "v201712.0037"),
    ("pycalver"       , "201712.38a0"        , None),
    ("pycalver"       , "201712.0039"        , None),
    ("semver"         , "1.23.456"           , "1.23.456"),
    ("calver"         , "v201712"            , "v201712"),
    ("calver"         , "v201799"            , None),  # invalid date
    ("calver"         , "201712"             , None),
    ("calver"         , "v20171"             , None),
    ("build"          , ".0012"              , ".0012"),
    ("build"          , ".11012"             , ".11012"),
    ("build"          , ".012"               , None),
    ("build"          , "11012"              , None),
    ("release"        , "-alpha"             , "-alpha"),
    ("release"        , "-beta"              , "-beta"),
    ("release"        , "-dev"               , "-dev"),
    ("release"        , "-rc"                , "-rc"),
    ("release"        , "-post"              , "-post"),
    ("release"        , "-pre"               , None),
    ("release"        , "alpha"              , None),
]


@pytest.mark.parametrize("part_name, line, expected", PATTERN_PART_CASES)
def test_re_pattern_parts(part_name, line, expected):
    part_re = _part_re_by_name(part_name)
    result  = part_re.search(line)
    if result is None:
        assert expected is None, (part_name, line)
    else:
        result_val = result.group(0)
        assert result_val == expected, (part_name, line)


CLI_MAIN_FIXTURE = """
@click.group()
@click.version_option(version="v201812.0123-beta")
@click.help_option()
"""


def test_pattern_escapes():
    pattern    = 'click.version_option(version="{pycalver}")'
    pattern_re = patterns.compile_pattern(pattern)
    match      = pattern_re.search(CLI_MAIN_FIXTURE)
    expected   = 'click.version_option(version="v201812.0123-beta")'
    assert match.group(0) == expected


CURLY_BRACE_FIXTURE = """
package_metadata = {"name": "mypackage", "version": "v201812.0123-beta"}
"""


def test_curly_escapes():
    pattern    = 'package_metadata = {"name": "mypackage", "version": "{pycalver}"}'
    pattern_re = patterns.compile_pattern(pattern)
    match      = pattern_re.search(CURLY_BRACE_FIXTURE)
    expected   = 'package_metadata = {"name": "mypackage", "version": "v201812.0123-beta"}'
    assert match.group(0) == expected



from pycalver import parse


SETUP_PY_FIXTURE = """
# setup.py
import setuptools
__version__ = 'v201712.0002-alpha'
setuptools.setup(
...
   version='201712.2a0',
"""


def test_default_parse_patterns():
    lines    = SETUP_PY_FIXTURE.splitlines()
    patterns = ["{pycalver}", "{pep440_pycalver}"]

    matches = list(parse.iter_matches(lines, patterns))
    assert len(matches) == 2

    assert matches[0].lineno == 3
    assert matches[1].lineno == 6

    assert matches[0].pattern == patterns[0]
    assert matches[1].pattern == patterns[1]

    assert matches[0].match == "v201712.0002-alpha"
    assert matches[1].match == "201712.2a0"


def test_explicit_parse_patterns():
    lines = SETUP_PY_FIXTURE.splitlines()

    patterns = ["__version__ = '{pycalver}'", "version='{pep440_pycalver}'"]

    matches = list(parse.iter_matches(lines, patterns))
    assert len(matches) == 2

    assert matches[0].lineno == 3
    assert matches[1].lineno == 6

    assert matches[0].pattern == patterns[0]
    assert matches[1].pattern == patterns[1]

    assert matches[0].match == "__version__ = 'v201712.0002-alpha'"
    assert matches[1].match == "version='201712.2a0'"


README_RST_FIXTURE = """
:alt: PyPI version

.. |version| image:: https://img.shields.io/badge/CalVer-v201809.0002--beta-blue.svg
:target: https://calver.org/
:alt: CalVer v201809.0002-beta
"""


def test_badge_parse_patterns():
    lines = README_RST_FIXTURE.splitlines()

    patterns = ["badge/CalVer-{calver}{build}-{release}-blue.svg", ":alt: CalVer {pycalver}"]

    matches = list(parse.iter_matches(lines, patterns))
    assert len(matches) == 2

    assert matches[0].lineno == 3
    assert matches[1].lineno == 5

    assert matches[0].pattern == patterns[0]
    assert matches[1].pattern == patterns[1]

    assert matches[0].match == "badge/CalVer-v201809.0002--beta-blue.svg"
    assert matches[1].match == ":alt: CalVer v201809.0002-beta"



# This file is part of the pycalver project
# https://gitlab.com/mbarkhau/pycalver
#
# Copyright (c) 2019 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT
"""PyCalVer: CalVer for Python Packages."""

__version__ = "v201903.0028"



import random
from pycalver import lex_id


def test_next_id_basic():
    assert lex_id.next_id("01") == "02"
    assert lex_id.next_id("09") == "110"


def test_next_id_overflow():
    try:
        prev_id = "9999"
        next_id = lex_id.next_id(prev_id)
        assert False, (prev_id, "->", next_id)
    except OverflowError:
        pass


def test_next_id_random():
    for i in range(1000):
        prev_id = str(random.randint(1, 100 * 1000))
        try:
            next_id = lex_id.next_id(prev_id)
            assert prev_id < next_id
        except OverflowError:
            assert len(prev_id) == prev_id.count("9")


def test_ord_val():
    assert lex_id.ord_val("1"  ) == 1
    assert lex_id.ord_val("01" ) == 1
    assert lex_id.ord_val("02" ) == 2
    assert lex_id.ord_val("09" ) == 9
    assert lex_id.ord_val("110") == 10


def test_main(capsys):
    lex_id._main()
    captured = capsys.readouterr()
    assert len(captured.err) == 0

    lines  = iter(captured.out.splitlines())
    header = next(lines)

    assert "lexical" in header
    assert "numerical" in header

    ids      = []
    ord_vals = []

    for line in lines:
        if "..." in line:
            continue
        _id, _ord_val = line.split()

        assert _id.endswith(_ord_val)
        assert int(_ord_val) == int(_ord_val, 10)

        ids.append(_id.strip())
        ord_vals.append(int(_ord_val.strip()))

    assert len(ids) > 0
    assert sorted(ids) == ids
    assert sorted(ord_vals) == ord_vals



import os
import time
import shutil
import pathlib2 as pl
import subprocess as sp

import pytest

from click.testing import CliRunner

import pycalver.config as config
import pycalver.patterns as patterns
import pycalver.__main__ as pycalver


SETUP_CFG_FIXTURE = """
[metadata]
license_file = LICENSE

[bdist_wheel]
universal = 1
"""

PYCALVER_TOML_FIXTURE = """
"""

PYPROJECT_TOML_FIXTURE = """
[build-system]
requires = ["setuptools", "wheel"]
"""

ENV = {
    'GIT_AUTHOR_NAME'    : "pycalver_tester",
    'GIT_COMMITTER_NAME' : "pycalver_tester",
    'GIT_AUTHOR_EMAIL'   : "pycalver_tester@nowhere.com",
    'GIT_COMMITTER_EMAIL': "pycalver_tester@nowhere.com",
    'HGUSER'             : "pycalver_tester",
}


def sh(*cmd):
    return sp.check_output(cmd, env=ENV)


@pytest.fixture
def runner(tmpdir):
    runner   = CliRunner(env=ENV)
    orig_cwd = os.getcwd()

    _debug = 0
    if _debug:
        tmpdir = pl.Path("..") / "tmp_test_pycalver_project"
        if tmpdir.exists():
            time.sleep(0.2)
            shutil.rmtree(str(tmpdir))
        tmpdir.mkdir()

    os.chdir(str(tmpdir))

    yield runner

    os.chdir(orig_cwd)

    if not _debug:
        shutil.rmtree(str(tmpdir))


def test_help(runner):
    result = runner.invoke(pycalver.cli, ['--help', "--verbose"])
    assert result.exit_code == 0
    assert "PyCalVer" in result.output
    assert "bump " in result.output
    assert "test " in result.output
    assert "init " in result.output
    assert "show " in result.output


def test_version(runner):
    result = runner.invoke(pycalver.cli, ['--version', "--verbose"])
    assert result.exit_code == 0
    assert " version v20" in result.output
    match = patterns.PYCALVER_RE.search(result.output)
    assert match


def test_incr_default(runner):
    old_version     = "v201701.0999-alpha"
    initial_version = config._initial_version()

    result = runner.invoke(pycalver.cli, ['test', "--verbose", old_version])
    assert result.exit_code == 0
    new_version = initial_version.replace(".0001-alpha", ".11000-alpha")
    assert f"Version: {new_version}\n" in result.output


def test_incr_semver(runner):
    semver_pattern = "{MAJOR}.{MINOR}.{PATCH}"
    old_version    = "0.1.0"
    new_version    = "0.1.1"

    result = runner.invoke(pycalver.cli, ['test', "--verbose", "--patch", old_version, "{semver}"])
    assert result.exit_code == 0
    assert f"Version: {new_version}\n" in result.output

    result = runner.invoke(
        pycalver.cli, ['test', "--verbose", "--patch", old_version, semver_pattern]
    )
    assert result.exit_code == 0
    assert f"Version: {new_version}\n" in result.output

    old_version = "0.1.1"
    new_version = "0.2.0"

    result = runner.invoke(
        pycalver.cli, ['test', "--verbose", "--minor", old_version, semver_pattern]
    )
    assert result.exit_code == 0
    assert f"Version: {new_version}\n" in result.output

    old_version = "0.1.1"
    new_version = "1.0.0"

    result = runner.invoke(
        pycalver.cli, ['test', "--verbose", "--major", old_version, semver_pattern]
    )
    assert result.exit_code == 0
    assert f"Version: {new_version}\n" in result.output


def test_incr_semver_invalid(runner, caplog):
    result = runner.invoke(pycalver.cli, ['test', "--verbose", "--patch", "0.1.1"])
    assert result.exit_code == 1
    assert len(caplog.records) > 0
    log_record = caplog.records[0]
    assert "Invalid version string" in log_record.message
    assert "for pattern '{pycalver}'" in log_record.message


def test_incr_to_beta(runner):
    old_version     = "v201701.0999-alpha"
    initial_version = config._initial_version()

    result = runner.invoke(pycalver.cli, ['test', old_version, "--verbose", "--release", "beta"])
    assert result.exit_code == 0
    new_version = initial_version.replace(".0001-alpha", ".11000-beta")
    assert f"Version: {new_version}\n" in result.output


def test_incr_to_final(runner):
    old_version     = "v201701.0999-alpha"
    initial_version = config._initial_version()

    result = runner.invoke(pycalver.cli, ['test', old_version, "--verbose", "--release", "final"])
    assert result.exit_code == 0
    new_version = initial_version.replace(".0001-alpha", ".11000")
    assert f"Version: {new_version}\n" in result.output


def test_incr_invalid(runner):
    old_version = "v201701.0999-alpha"

    result = runner.invoke(pycalver.cli, ['test', old_version, "--verbose", "--release", "alfa"])
    assert result.exit_code == 1


def _add_project_files(*files):
    if "README.md" in files:
        with pl.Path("README.md").open(mode="wt", encoding="utf-8") as fh:
            fh.write(
                """
                Hello World v201701.0002-alpha !
                aka. 201701.2a0 !
            """
            )

    if "setup.cfg" in files:
        with pl.Path("setup.cfg").open(mode="wt", encoding="utf-8") as fh:
            fh.write(SETUP_CFG_FIXTURE)

    if "pycalver.toml" in files:
        with pl.Path("pycalver.toml").open(mode="wt", encoding="utf-8") as fh:
            fh.write(PYCALVER_TOML_FIXTURE)

    if "pyproject.toml" in files:
        with pl.Path("pyproject.toml").open(mode="wt", encoding="utf-8") as fh:
            fh.write(PYPROJECT_TOML_FIXTURE)


def test_nocfg(runner, caplog):
    _add_project_files("README.md")
    result = runner.invoke(pycalver.cli, ['show', "--verbose"])
    assert result.exit_code == 1
    assert any(
        bool("Could not parse configuration. Perhaps try 'pycalver init'." in r.message)
        for r in caplog.records
    )


def test_novcs_nocfg_init(runner, caplog, capsys):
    _add_project_files("README.md")
    # dry mode test
    result = runner.invoke(pycalver.cli, ['init', "--verbose", "--dry"])
    assert result.exit_code == 0
    assert not os.path.exists("pycalver.toml")

    # check logging
    assert len(caplog.records) == 1
    log = caplog.records[0]
    assert log.levelname == 'WARNING'
    assert "File not found" in log.message

    # print("moep")
    # captured = capsys.readouterr()
    # assert not captured.err
    # assert "Would have written to pycalver.toml:" in captured.out

    # non dry mode
    result = runner.invoke(pycalver.cli, ['init', "--verbose"])
    assert result.exit_code == 0

    # check logging
    assert len(caplog.records) == 2
    log = caplog.records[1]
    assert log.levelname == 'WARNING'
    assert "File not found" in log.message

    assert os.path.exists("pycalver.toml")
    with pl.Path("pycalver.toml").open(mode="r", encoding="utf-8") as fh:
        cfg_content = fh.read()

    base_str = config.DEFAULT_TOML_BASE_TMPL.format(initial_version=config._initial_version())
    assert base_str                          in cfg_content
    assert config.DEFAULT_TOML_README_MD_STR in cfg_content

    result = runner.invoke(pycalver.cli, ['show', "--verbose"])
    assert result.exit_code == 0
    assert f"Current Version: {config._initial_version()}\n" in result.output
    assert f"PEP440         : {config._initial_version_pep440()}\n" in result.output

    result = runner.invoke(pycalver.cli, ['init', "--verbose"])
    assert result.exit_code == 1

    # check logging
    assert len(caplog.records) == 3
    log = caplog.records[2]
    assert log.levelname == 'ERROR'
    assert "Configuration already initialized" in log.message


def test_novcs_setupcfg_init(runner):
    _add_project_files("README.md", "setup.cfg")
    result = runner.invoke(pycalver.cli, ['init', "--verbose"])
    assert result.exit_code == 0

    with pl.Path("setup.cfg").open(mode="r", encoding="utf-8") as fh:
        cfg_content = fh.read()

    base_str = config.DEFAULT_CONFIGPARSER_BASE_TMPL.format(
        initial_version=config._initial_version()
    )
    assert base_str                                  in cfg_content
    assert config.DEFAULT_CONFIGPARSER_README_MD_STR in cfg_content

    result = runner.invoke(pycalver.cli, ['show', "--verbose"])
    assert result.exit_code == 0
    assert f"Current Version: {config._initial_version()}\n" in result.output
    assert f"PEP440         : {config._initial_version_pep440()}\n" in result.output


def test_novcs_pyproject_init(runner):
    _add_project_files("README.md", "pyproject.toml")
    result = runner.invoke(pycalver.cli, ['init', "--verbose"])
    assert result.exit_code == 0

    with pl.Path("pyproject.toml").open(mode="r", encoding="utf-8") as fh:
        cfg_content = fh.read()

    base_str = config.DEFAULT_TOML_BASE_TMPL.format(initial_version=config._initial_version())
    assert base_str                          in cfg_content
    assert config.DEFAULT_TOML_README_MD_STR in cfg_content

    result = runner.invoke(pycalver.cli, ['show'])
    assert result.exit_code == 0
    assert f"Current Version: {config._initial_version()}\n" in result.output
    assert f"PEP440         : {config._initial_version_pep440()}\n" in result.output


def _vcs_init(vcs):
    assert vcs in ("git", "hg")
    assert not pl.Path(f".{vcs}").exists()
    sh(f"{vcs}", "init")
    assert pl.Path(f".{vcs}").is_dir()

    sh(f"{vcs}", "add", "README.md")
    sh(f"{vcs}", "commit", "-m", "initial commit")


def test_git_init(runner):
    _add_project_files("README.md")
    _vcs_init("git")

    result = runner.invoke(pycalver.cli, ['init', "--verbose"])
    assert result.exit_code == 0

    result = runner.invoke(pycalver.cli, ['show'])
    assert result.exit_code == 0
    assert f"Current Version: {config._initial_version()}\n" in result.output
    assert f"PEP440         : {config._initial_version_pep440()}\n" in result.output


def test_hg_init(runner):
    _add_project_files("README.md")
    _vcs_init("hg")

    result = runner.invoke(pycalver.cli, ['init', "--verbose"])
    assert result.exit_code == 0

    result = runner.invoke(pycalver.cli, ['show'])
    assert result.exit_code == 0
    assert f"Current Version: {config._initial_version()}\n" in result.output
    assert f"PEP440         : {config._initial_version_pep440()}\n" in result.output


def test_git_tag_eval(runner):
    _add_project_files("README.md")
    _vcs_init("git")

    # This will set a version that is older than the version tag
    # we set in the vcs, which should take precedence.
    result = runner.invoke(pycalver.cli, ['init', "--verbose"])
    assert result.exit_code == 0
    initial_version    = config._initial_version()
    tag_version        = initial_version.replace(".0001-alpha", ".0123-beta")
    tag_version_pep440 = tag_version[1:7] + ".123b0"

    sh("git", "tag", "--annotate", tag_version, "--message", f"bump version to {tag_version}")

    result = runner.invoke(pycalver.cli, ['show', "--verbose"])
    assert result.exit_code == 0
    assert f"Current Version: {tag_version}\n" in result.output
    assert f"PEP440         : {tag_version_pep440}\n" in result.output


def test_hg_tag_eval(runner):
    _add_project_files("README.md")
    _vcs_init("hg")

    # This will set a version that is older than the version tag
    # we set in the vcs, which should take precedence.
    result = runner.invoke(pycalver.cli, ['init', "--verbose"])
    assert result.exit_code == 0
    initial_version    = config._initial_version()
    tag_version        = initial_version.replace(".0001-alpha", ".0123-beta")
    tag_version_pep440 = tag_version[1:7] + ".123b0"

    sh("hg", "tag", tag_version, "--message", f"bump version to {tag_version}")

    result = runner.invoke(pycalver.cli, ['show', "--verbose"])
    assert result.exit_code == 0
    assert f"Current Version: {tag_version}\n" in result.output
    assert f"PEP440         : {tag_version_pep440}\n" in result.output


def test_novcs_bump(runner):
    _add_project_files("README.md")

    result = runner.invoke(pycalver.cli, ['init', "--verbose"])
    assert result.exit_code == 0

    result = runner.invoke(pycalver.cli, ['bump', "--verbose"])
    assert result.exit_code == 0

    calver = config._initial_version()[:7]

    with pl.Path("README.md").open() as fh:
        content = fh.read()
        assert calver + ".0002-alpha !\n" in content
        assert calver[1:] + ".2a0 !\n" in content

    result = runner.invoke(pycalver.cli, ['bump', "--verbose", "--release", "beta"])
    assert result.exit_code == 0

    with pl.Path("README.md").open() as fh:
        content = fh.read()
        assert calver + ".0003-beta !\n" in content
        assert calver[1:] + ".3b0 !\n" in content


def test_git_bump(runner):
    _add_project_files("README.md")
    _vcs_init("git")

    result = runner.invoke(pycalver.cli, ['init', "--verbose"])
    assert result.exit_code == 0

    sh("git", "add", "pycalver.toml")
    sh("git", "commit", "-m", "initial commit")

    result = runner.invoke(pycalver.cli, ['bump', "--verbose"])
    assert result.exit_code == 0

    calver = config._initial_version()[:7]

    with pl.Path("README.md").open() as fh:
        content = fh.read()
        assert calver + ".0002-alpha !\n" in content


def test_hg_bump(runner):
    _add_project_files("README.md")
    _vcs_init("hg")

    result = runner.invoke(pycalver.cli, ['init', "--verbose"])
    assert result.exit_code == 0

    sh("hg", "add", "pycalver.toml")
    sh("hg", "commit", "-m", "initial commit")

    result = runner.invoke(pycalver.cli, ['bump', "--verbose"])
    assert result.exit_code == 0

    calver = config._initial_version()[:7]

    with pl.Path("README.md").open() as fh:
        content = fh.read()
        assert calver + ".0002-alpha !\n" in content


def test_empty_git_bump(runner, caplog):
    sh("git", "init")
    with pl.Path("setup.cfg").open(mode="w") as fh:
        fh.write("")
    result = runner.invoke(pycalver.cli, ['init', "--verbose"])
    assert result.exit_code == 0

    with pl.Path("setup.cfg").open(mode="r") as fh:
        default_cfg_data = fh.read()

    assert "[pycalver]\n" in default_cfg_data
    assert "\ncurrent_version = " in default_cfg_data
    assert "\n[pycalver:file_patterns]\n" in default_cfg_data
    assert "\nsetup.cfg =\n" in default_cfg_data

    result = runner.invoke(pycalver.cli, ['bump'])

    assert any(("working directory is not clean" in r.message) for r in caplog.records)
    assert any(("setup.cfg" in r.message) for r in caplog.records)


def test_empty_hg_bump(runner, caplog):
    sh("hg", "init")
    with pl.Path("setup.cfg").open(mode="w") as fh:
        fh.write("")
    result = runner.invoke(pycalver.cli, ['init', "--verbose"])
    assert result.exit_code == 0

    with pl.Path("setup.cfg").open(mode="r") as fh:
        default_cfg_data = fh.read()

    assert "[pycalver]\n" in default_cfg_data
    assert "\ncurrent_version = " in default_cfg_data
    assert "\n[pycalver:file_patterns]\n" in default_cfg_data
    assert "\nsetup.cfg =\n" in default_cfg_data

    result = runner.invoke(pycalver.cli, ['bump'])

    assert any(("working directory is not clean" in r.message) for r in caplog.records)
    assert any(("setup.cfg" in r.message) for r in caplog.records)



# This file is part of the pycalver project
# https://gitlab.com/mbarkhau/pycalver
#
# Copyright (c) 2019 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT
"""Parse setup.cfg or pycalver.cfg files."""

import io
import os
import six
import toml
import configparser
import typing as typ
import pathlib2 as pl
import datetime as dt

import logging

from . import version

log = logging.getLogger("pycalver.config")

Patterns       = typ.List[str]
PatternsByGlob = typ.Dict[str, Patterns]

SUPPORTED_CONFIGS = ["setup.cfg", "pyproject.toml", "pycalver.toml"]


class ProjectContext(typ.NamedTuple):
    """Container class for project info."""

    path           : pl.Path
    config_filepath: pl.Path
    config_format  : str
    vcs_type       : typ.Optional[str]


def init_project_ctx(project_path: typ.Union[str, pl.Path, None] = ".") -> ProjectContext:
    """Initialize ProjectContext from a path."""
    if isinstance(project_path, pl.Path):
        path = project_path
    elif project_path is None:
        path = pl.Path(".")
    else:
        # assume it's a str/unicode
        path = pl.Path(project_path)

    if (path / "pycalver.toml").exists():
        config_filepath = path / "pycalver.toml"
        config_format   = 'toml'
    elif (path / "pyproject.toml").exists():
        config_filepath = path / "pyproject.toml"
        config_format   = 'toml'
    elif (path / "setup.cfg").exists():
        config_filepath = path / "setup.cfg"
        config_format   = 'cfg'
    else:
        # fallback to creating a new pycalver.toml
        config_filepath = path / "pycalver.toml"
        config_format   = 'toml'

    vcs_type: typ.Optional[str]

    if (path / ".git").exists():
        vcs_type = 'git'
    elif (path / ".hg").exists():
        vcs_type = 'hg'
    else:
        vcs_type = None

    return ProjectContext(path, config_filepath, config_format, vcs_type)


RawConfig = typ.Dict[str, typ.Any]


class Config(typ.NamedTuple):
    """Container for parameters parsed from a config file."""

    current_version: str
    version_pattern: str
    pep440_version : str

    commit: bool
    tag   : bool
    push  : bool

    file_patterns: PatternsByGlob


def _debug_str(cfg: Config) -> str:
    cfg_str_parts = [
        f"Config Parsed: Config(",
        f"current_version='{cfg.current_version}'",
        "version_pattern='{pycalver}'",
        f"pep440_version='{cfg.pep440_version}'",
        f"commit={cfg.commit}",
        f"tag={cfg.tag}",
        f"push={cfg.push}",
        "file_patterns={",
    ]

    for filepath, patterns in cfg.file_patterns.items():
        for pattern in patterns:
            cfg_str_parts.append(f"\n    '{filepath}': '{pattern}'")

    cfg_str_parts += ["\n})"]
    return ", ".join(cfg_str_parts)


MaybeConfig    = typ.Optional[Config]
MaybeRawConfig = typ.Optional[RawConfig]

FilePatterns = typ.Dict[str, typ.List[str]]


def _parse_cfg_file_patterns(cfg_parser: configparser.RawConfigParser,) -> FilePatterns:

    file_patterns: FilePatterns = {}

    for filepath, patterns_str in cfg_parser.items("pycalver:file_patterns"):
        patterns: typ.List[str] = []
        for line in patterns_str.splitlines():
            pattern = line.strip()
            if pattern:
                patterns.append(pattern)

        file_patterns[filepath] = patterns

    return file_patterns


class _ConfigParser(configparser.RawConfigParser):
    """Custom parser, simply to override optionxform behaviour."""

    def optionxform(self, optionstr: str) -> str:
        """Non-xforming (ie. uppercase preserving) override.

        This is important because our option names are actually
        filenames, so case sensitivity is relevant. The default
        behaviour is to do optionstr.lower()
        """
        return optionstr


OptionVal = typ.Union[str, bool, None]

BOOL_OPTIONS: typ.Mapping[str, OptionVal] = {'commit': False, 'tag': None, 'push': None}


def _parse_cfg(cfg_buffer: typ.IO[str]) -> RawConfig:
    cfg_parser = _ConfigParser()

    if hasattr(cfg_parser, 'read_file'):
        cfg_parser.read_file(cfg_buffer)
    else:
        cfg_parser.readfp(cfg_buffer)  # python2 compat

    if not cfg_parser.has_section("pycalver"):
        raise ValueError("Missing [pycalver] section.")

    raw_cfg: RawConfig = dict(cfg_parser.items("pycalver"))

    for option, default_val in BOOL_OPTIONS.items():
        val: OptionVal = raw_cfg.get(option, default_val)
        if isinstance(val, six.text_type):
            val = val.lower() in ("yes", "true", "1", "on")
        raw_cfg[option] = val

    raw_cfg['file_patterns'] = _parse_cfg_file_patterns(cfg_parser)

    return raw_cfg


def _parse_toml(cfg_buffer: typ.IO[str]) -> RawConfig:
    raw_full_cfg = toml.load(cfg_buffer)
    raw_cfg      = raw_full_cfg.get('pycalver', {})

    for option, default_val in BOOL_OPTIONS.items():
        raw_cfg[option] = raw_cfg.get(option, default_val)

    return raw_cfg


def _normalize_file_patterns(raw_cfg: RawConfig) -> FilePatterns:
    """Create consistent representation of file_patterns.

    The result the same, regardless of the config format.
    """
    version_str    : str = raw_cfg['current_version']
    version_pattern: str = raw_cfg['version_pattern']
    pep440_version : str = version.to_pep440(version_str)

    file_patterns: FilePatterns = raw_cfg['file_patterns']

    for filepath, patterns in list(file_patterns.items()):
        if not os.path.exists(filepath):
            log.warning(f"Invalid config, no such file: {filepath}")

        normalized_patterns: typ.List[str] = []
        for pattern in patterns:
            normalized_pattern = pattern.replace("{version}", version_pattern)
            if version_pattern == "{pycalver}":
                normalized_pattern = normalized_pattern.replace(
                    "{pep440_version}", "{pep440_pycalver}"
                )
            elif version_pattern == "{semver}":
                normalized_pattern = normalized_pattern.replace("{pep440_version}", "{semver}")
            elif "{pep440_version}" in pattern:
                log.warning(f"Invalid config, cannot match '{pattern}' for '{filepath}'.")
                log.warning(f"No mapping of '{version_pattern}' to '{pep440_version}'")
            normalized_patterns.append(normalized_pattern)

        file_patterns[filepath] = normalized_patterns

    return file_patterns


def _parse_config(raw_cfg: RawConfig) -> Config:
    """Parse configuration which was loaded from an .ini/.cfg or .toml file."""

    if 'current_version' not in raw_cfg:
        raise ValueError("Missing 'pycalver.current_version'")

    version_str: str = raw_cfg['current_version']
    version_str = raw_cfg['current_version'] = version_str.strip("'\" ")

    version_pattern: str = raw_cfg.get('version_pattern', "{pycalver}")
    version_pattern = raw_cfg['version_pattern'] = version_pattern.strip("'\" ")

    # NOTE (mb 2019-01-05): Provoke ValueError if version_pattern
    #   and current_version are not compatible.
    version.parse_version_info(version_str, version_pattern)

    pep440_version = version.to_pep440(version_str)

    commit = raw_cfg['commit']
    tag    = raw_cfg['tag']
    push   = raw_cfg['push']

    if tag is None:
        tag = raw_cfg['tag'] = False
    if push is None:
        push = raw_cfg['push'] = False

    if tag and not commit:
        raise ValueError("pycalver.commit = true required if pycalver.tag = true")

    if push and not commit:
        raise ValueError("pycalver.commit = true required if pycalver.push = true")

    file_patterns = _normalize_file_patterns(raw_cfg)

    cfg = Config(
        current_version=version_str,
        version_pattern=version_pattern,
        pep440_version=pep440_version,
        commit=commit,
        tag=tag,
        push=push,
        file_patterns=file_patterns,
    )
    log.debug(_debug_str(cfg))
    return cfg


def parse(ctx: ProjectContext) -> MaybeConfig:
    """Parse config file if available."""
    if not ctx.config_filepath.exists():
        log.warning(f"File not found: {ctx.config_filepath}")
        return None

    fh: typ.IO[str]

    try:
        with ctx.config_filepath.open(mode="rt", encoding="utf-8") as fh:
            if ctx.config_format == 'toml':
                raw_cfg = _parse_toml(fh)
            elif ctx.config_format == 'cfg':
                raw_cfg = _parse_cfg(fh)
            else:
                err_msg = "Invalid config_format='{ctx.config_format}'"
                raise RuntimeError(err_msg)

        return _parse_config(raw_cfg)
    except ValueError as ex:
        log.warning(f"Couldn't parse {ctx.config_filepath}: {str(ex)}")
        return None


DEFAULT_CONFIGPARSER_BASE_TMPL = """
[pycalver]
current_version = "{initial_version}"
version_pattern = "{{pycalver}}"
commit = True
tag = True
push = True

[pycalver:file_patterns]
""".lstrip()


DEFAULT_CONFIGPARSER_SETUP_CFG_STR = """
setup.cfg =
    current_version = "{version}"
""".lstrip()


DEFAULT_CONFIGPARSER_SETUP_PY_STR = """
setup.py =
    "{version}"
    "{pep440_version}"
""".lstrip()


DEFAULT_CONFIGPARSER_README_RST_STR = """
README.rst =
    {version}
    {pep440_version}
""".lstrip()


DEFAULT_CONFIGPARSER_README_MD_STR = """
README.md =
    {version}
    {pep440_version}
""".lstrip()


DEFAULT_TOML_BASE_TMPL = """
[pycalver]
current_version = "{initial_version}"
version_pattern = "{{pycalver}}"
commit = true
tag = true
push = true

[pycalver.file_patterns]
""".lstrip()


DEFAULT_TOML_PYCALVER_STR = """
"pycalver.toml" = [
    'current_version = "{version}"',
]
""".lstrip()


DEFAULT_TOML_PYPROJECT_STR = """
"pyproject.toml" = [
    'current_version = "{version}"',
]
""".lstrip()


DEFAULT_TOML_SETUP_PY_STR = """
"setup.py" = [
    "{version}",
    "{pep440_version}",
]
""".lstrip()


DEFAULT_TOML_README_RST_STR = """
"README.rst" = [
    "{version}",
    "{pep440_version}",
]
""".lstrip()


DEFAULT_TOML_README_MD_STR = """
"README.md" = [
    "{version}",
    "{pep440_version}",
]
""".lstrip()


def _initial_version() -> str:
    return dt.datetime.now().strftime("v%Y%m.0001-alpha")


def _initial_version_pep440() -> str:
    return dt.datetime.now().strftime("%Y%m.1a0")


def default_config(ctx: ProjectContext) -> str:
    """Generate initial default config."""
    fmt = ctx.config_format
    if fmt == 'cfg':
        base_tmpl = DEFAULT_CONFIGPARSER_BASE_TMPL

        default_pattern_strs_by_filename = {
            "setup.cfg" : DEFAULT_CONFIGPARSER_SETUP_CFG_STR,
            "setup.py"  : DEFAULT_CONFIGPARSER_SETUP_PY_STR,
            "README.rst": DEFAULT_CONFIGPARSER_README_RST_STR,
            "README.md" : DEFAULT_CONFIGPARSER_README_MD_STR,
        }
    elif fmt == 'toml':
        base_tmpl = DEFAULT_TOML_BASE_TMPL

        default_pattern_strs_by_filename = {
            "pyproject.toml": DEFAULT_TOML_PYPROJECT_STR,
            "pycalver.toml" : DEFAULT_TOML_PYCALVER_STR,
            "setup.py"      : DEFAULT_TOML_SETUP_PY_STR,
            "README.rst"    : DEFAULT_TOML_README_RST_STR,
            "README.md"     : DEFAULT_TOML_README_MD_STR,
        }
    else:
        raise ValueError(f"Invalid config_format='{fmt}', must be either 'toml' or 'cfg'.")

    cfg_str = base_tmpl.format(initial_version=_initial_version())

    for filename, default_str in default_pattern_strs_by_filename.items():
        if (ctx.path / filename).exists():
            cfg_str += default_str

    has_config_file = any((ctx.path / fn).exists() for fn in SUPPORTED_CONFIGS)

    if not has_config_file:
        if ctx.config_format == 'cfg':
            cfg_str += DEFAULT_CONFIGPARSER_SETUP_CFG_STR
        if ctx.config_format == 'toml':
            cfg_str += DEFAULT_TOML_PYCALVER_STR

    cfg_str += "\n"

    return cfg_str


def write_content(ctx: ProjectContext) -> None:
    """Update project config file with initial default config."""
    # path           : pl.Path
    # config_filepath: pl.Path
    # config_format  : str
    # vcs_type       : typ.Optional[str]

    fh: typ.IO[str]

    cfg_content = default_config(ctx)
    if os.path.exists("pyproject.toml"):
        with io.open("pyproject.toml", mode="at", encoding="utf-8") as fh:
            fh.write(cfg_content)
        print("Updated pyproject.toml")
    elif os.path.exists("setup.cfg"):
        with io.open("setup.cfg", mode="at", encoding="utf-8") as fh:
            fh.write(cfg_content)
        print("Updated setup.cfg")
    else:
        with io.open("pycalver.toml", mode="at", encoding="utf-8") as fh:
            fh.write(cfg_content)
        print("Created pycalver.toml")



# This file is part of the pycalver project
# https://gitlab.com/mbarkhau/pycalver
#
# Copyright (c) 2019 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT
"""Compose Regular Expressions from Patterns.

>>> version_info = PYCALVER_RE.match("v201712.0123-alpha").groupdict()
>>> assert version_info == {
...     "pycalver"    : "v201712.0123-alpha",
...     "vYYYYMM"     : "v201712",
...     "year"        : "2017",
...     "month"       : "12",
...     "build"       : ".0123",
...     "build_no"    : "0123",
...     "release"     : "-alpha",
...     "release_tag" : "alpha",
... }
>>>
>>> version_info = PYCALVER_RE.match("v201712.0033").groupdict()
>>> assert version_info == {
...     "pycalver"   : "v201712.0033",
...     "vYYYYMM"    : "v201712",
...     "year"       : "2017",
...     "month"      : "12",
...     "build"      : ".0033",
...     "build_no"   : "0033",
...     "release"    : None,
...     "release_tag": None,
... }
"""

import re
import typing as typ

# https://regex101.com/r/fnj60p/10
PYCALVER_PATTERN = r"""
\b
(?P<pycalver>
    (?P<vYYYYMM>
       v                        # "v" version prefix
       (?P<year>\d{4})
       (?P<month>\d{2})
    )
    (?P<build>
        \.                      # "." build nr prefix
        (?P<build_no>\d{4,})
    )
    (?P<release>
        \-                      # "-" release prefix
        (?P<release_tag>alpha|beta|dev|rc|post)
    )?
)(?:\s|$)
"""

PYCALVER_RE: typ.Pattern[str] = re.compile(PYCALVER_PATTERN, flags=re.VERBOSE)


PATTERN_ESCAPES = [
    ("\u005c", "\u005c\u005c"),
    ("-"     , "\u005c-"),
    ("."     , "\u005c."),
    ("+"     , "\u005c+"),
    ("*"     , "\u005c*"),
    ("?"     , "\u005c?"),
    ("{"     , "\u005c{"),
    ("}"     , "\u005c}"),
    ("["     , "\u005c["),
    ("]"     , "\u005c]"),
    ("("     , "\u005c("),
    (")"     , "\u005c)"),
]

COMPOSITE_PART_PATTERNS = {
    'pep440_pycalver': r"{year}{month}\.{BID}(?:{pep440_tag})?",
    'pycalver'       : r"v{year}{month}\.{bid}(?:-{tag})?",
    'calver'         : r"v{year}{month}",
    'semver'         : r"{MAJOR}\.{MINOR}\.{PATCH}",
    'release_tag'    : r"{tag}",
    'build'          : r"\.{bid}",
    'release'        : r"-{tag}",
    # depricated
    'pep440_version': r"{year}{month}\.{BID}(?:{pep440_tag})?",
}


PART_PATTERNS = {
    'year'      : r"\d{4}",
    'month'     : r"(?:0[0-9]|1[0-2])",
    'build_no'  : r"\d{4,}",
    'pep440_tag': r"(?:a|b|dev|rc|post)?\d*",
    'tag'       : r"(?:alpha|beta|dev|rc|post|final)",
    'yy'        : r"\d{2}",
    'yyyy'      : r"\d{4}",
    'quarter'   : r"[1-4]",
    'iso_week'  : r"(?:[0-4]\d|5[0-3])",
    'us_week'   : r"(?:[0-4]\d|5[0-3])",
    'dom'       : r"(0[1-9]|[1-2][0-9]|3[0-1])",
    'doy'       : r"(?:[0-2]\d\d|3[0-5][0-9]|36[0-6])",
    'MAJOR'     : r"\d+",
    'MINOR'     : r"\d+",
    'MM'        : r"\d{2,}",
    'MMM'       : r"\d{3,}",
    'MMMM'      : r"\d{4,}",
    'MMMMM'     : r"\d{5,}",
    'PATCH'     : r"\d+",
    'PP'        : r"\d{2,}",
    'PPP'       : r"\d{3,}",
    'PPPP'      : r"\d{4,}",
    'PPPPP'     : r"\d{5,}",
    'bid'       : r"\d{4,}",
    'BID'       : r"[1-9]\d*",
    'BB'        : r"[1-9]\d{1,}",
    'BBB'       : r"[1-9]\d{2,}",
    'BBBB'      : r"[1-9]\d{3,}",
    'BBBBB'     : r"[1-9]\d{4,}",
    'BBBBBB'    : r"[1-9]\d{5,}",
    'BBBBBBB'   : r"[1-9]\d{6,}",
}


FULL_PART_FORMATS = {
    'pep440_pycalver': "{year}{month:02}.{BID}{pep440_tag}",
    'pycalver'       : "v{year}{month:02}.{bid}{release}",
    'calver'         : "v{year}{month:02}",
    'semver'         : "{MAJOR}.{MINOR}.{PATCH}",
    'release_tag'    : "{tag}",
    'build'          : ".{bid}",
    # NOTE (mb 2019-01-04): since release is optional, it
    # is treates specially in version.format
    # 'release'       : "-{tag}",
    'month'   : "{month:02}",
    'build_no': "{bid}",
    'iso_week': "{iso_week:02}",
    'us_week' : "{us_week:02}",
    'dom'     : "{dom:02}",
    'doy'     : "{doy:03}",
    # depricated
    'pep440_version': "{year}{month:02}.{BID}{pep440_tag}",
    'version'       : "v{year}{month:02}.{bid}{release}",
}


PART_FORMATS = {
    'major'  : "[0-9]+",
    'minor'  : "[0-9]{3,}",
    'patch'  : "[0-9]{3,}",
    'bid'    : "[0-9]{4,}",
    'MAJOR'  : "[0-9]+",
    'MINOR'  : "[0-9]+",
    'MM'     : "[0-9]{2,}",
    'MMM'    : "[0-9]{3,}",
    'MMMM'   : "[0-9]{4,}",
    'MMMMM'  : "[0-9]{5,}",
    'MMMMMM' : "[0-9]{6,}",
    'MMMMMMM': "[0-9]{7,}",
    'PATCH'  : "[0-9]+",
    'PP'     : "[0-9]{2,}",
    'PPP'    : "[0-9]{3,}",
    'PPPP'   : "[0-9]{4,}",
    'PPPPP'  : "[0-9]{5,}",
    'PPPPPP' : "[0-9]{6,}",
    'PPPPPPP': "[0-9]{7,}",
    'BID'    : "[1-9][0-9]*",
    'BB'     : "[1-9][0-9]{1,}",
    'BBB'    : "[1-9][0-9]{2,}",
    'BBBB'   : "[1-9][0-9]{3,}",
    'BBBBB'  : "[1-9][0-9]{4,}",
    'BBBBBB' : "[1-9][0-9]{5,}",
    'BBBBBBB': "[1-9][0-9]{6,}",
}


def _replace_pattern_parts(pattern: str) -> str:
    for part_name, part_pattern in PART_PATTERNS.items():
        named_part_pattern = f"(?P<{part_name}>{part_pattern})"
        placeholder        = "\u005c{" + part_name + "\u005c}"
        pattern            = pattern.replace(placeholder, named_part_pattern)
    return pattern


def _compile_pattern(pattern: str) -> str:
    for char, escaped in PATTERN_ESCAPES:
        pattern = pattern.replace(char, escaped)

    return _replace_pattern_parts(pattern)


def compile_pattern(pattern: str) -> typ.Pattern[str]:
    pattern_str = _compile_pattern(pattern)
    return re.compile(pattern_str)


def _init_composite_patterns() -> None:
    for part_name, part_pattern in COMPOSITE_PART_PATTERNS.items():
        part_pattern = part_pattern.replace("{", "\u005c{").replace("}", "\u005c}")
        pattern_str  = _replace_pattern_parts(part_pattern)
        PART_PATTERNS[part_name] = pattern_str


_init_composite_patterns()



# This file is part of the pycalver project
# https://gitlab.com/mbarkhau/pycalver
#
# Copyright (c) 2019 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT
"""Rewrite files, updating occurences of version strings."""

import io
import glob
import difflib
import logging
import typing as typ
import pathlib2 as pl

from . import parse
from . import config
from . import version
from . import patterns


log = logging.getLogger("pycalver.rewrite")


def detect_line_sep(content: str) -> str:
    r"""Parse line separator from content.

    >>> detect_line_sep('\r\n')
    '\r\n'
    >>> detect_line_sep('\r')
    '\r'
    >>> detect_line_sep('\n')
    '\n'
    >>> detect_line_sep('')
    '\n'
    """
    if "\r\n" in content:
        return "\r\n"
    elif "\r" in content:
        return "\r"
    else:
        return "\n"


def rewrite_lines(
    pattern_strs: typ.List[str], new_version: str, old_lines: typ.List[str]
) -> typ.List[str]:
    """Replace occurances of pattern_strs in old_lines with new_version.

    >>> pattern_strs = ['__version__ = "{pycalver}"']
    >>> rewrite_lines(pattern_strs, "v201811.0123-beta", ['__version__ = "v201809.0002-beta"'])
    ['__version__ = "v201811.0123-beta"']

    >>> pattern_strs = ['__version__ = "{pep440_version}"']
    >>> rewrite_lines(pattern_strs, "v201811.0123-beta", ['__version__ = "201809.2b0"'])
    ['__version__ = "201811.123b0"']
    """
    new_version_nfo = version.parse_version_info(new_version)

    new_lines      = old_lines[:]
    found_patterns = set()

    for m in parse.iter_matches(old_lines, pattern_strs):
        found_patterns.add(m.pattern)
        replacement = version.format_version(new_version_nfo, m.pattern)
        span_l, span_r = m.span
        new_line = m.line[:span_l] + replacement + m.line[span_r:]
        new_lines[m.lineno] = new_line

    non_matched_patterns = set(pattern_strs) - found_patterns
    if non_matched_patterns:
        for non_matched_pattern in non_matched_patterns:
            log.error(f"No match for pattern '{non_matched_pattern}'")
            compiled_pattern = patterns._compile_pattern(non_matched_pattern)
            log.error(f"Pattern compiles to regex '{compiled_pattern}'")
        raise ValueError("Invalid pattern(s)")
    else:
        return new_lines


class RewrittenFileData(typ.NamedTuple):
    """Container for line-wise content of rewritten files."""

    path     : str
    line_sep : str
    old_lines: typ.List[str]
    new_lines: typ.List[str]


def rfd_from_content(
    pattern_strs: typ.List[str], new_version: str, content: str
) -> RewrittenFileData:
    r"""Rewrite pattern occurrences with version string.

    >>> pattern_strs = ['__version__ = "{pycalver}"']
    >>> content = '__version__ = "v201809.0001-alpha"'
    >>> rfd = rfd_from_content(pattern_strs, "v201809.0123", content)
    >>> rfd.new_lines
    ['__version__ = "v201809.0123"']
    """
    line_sep  = detect_line_sep(content)
    old_lines = content.split(line_sep)
    new_lines = rewrite_lines(pattern_strs, new_version, old_lines)
    return RewrittenFileData("<path>", line_sep, old_lines, new_lines)


def _iter_file_paths(
    file_patterns: config.PatternsByGlob
) -> typ.Iterable[typ.Tuple[pl.Path, config.Patterns]]:
    for globstr, pattern_strs in file_patterns.items():
        file_paths = glob.glob(globstr)
        if not any(file_paths):
            errmsg = f"No files found for path/glob '{globstr}'"
            raise ValueError(errmsg)
        for file_path_str in file_paths:
            file_path = pl.Path(file_path_str)
            yield (file_path, pattern_strs)


def iter_rewritten(
    file_patterns: config.PatternsByGlob, new_version: str
) -> typ.Iterable[RewrittenFileData]:
    r'''Iterate over files with version string replaced.

    >>> file_patterns = {"src/pycalver/__init__.py": ['__version__ = "{pycalver}"']}
    >>> rewritten_datas = iter_rewritten(file_patterns, "v201809.0123")
    >>> rfd = list(rewritten_datas)[0]
    >>> assert rfd.new_lines == [
    ...     '# This file is part of the pycalver project',
    ...     '# https://gitlab.com/mbarkhau/pycalver',
    ...     '#',
    ...     '# Copyright (c) 2019 Manuel Barkhau (mbarkhau@gmail.com) - MIT License',
    ...     '# SPDX-License-Identifier: MIT',
    ...     '"""PyCalVer: CalVer for Python Packages."""',
    ...     '',
    ...     '__version__ = "v201809.0123"',
    ...     '',
    ... ]
    >>>
    '''
    fh: typ.IO[str]

    for file_path, pattern_strs in _iter_file_paths(file_patterns):
        with file_path.open(mode="rt", encoding="utf-8") as fh:
            content = fh.read()

        rfd = rfd_from_content(pattern_strs, new_version, content)
        yield rfd._replace(path=str(file_path))


def diff_lines(rfd: RewrittenFileData) -> typ.List[str]:
    r"""Generate unified diff.

    >>> rfd = RewrittenFileData(
    ...    path      = "<path>",
    ...    line_sep  = "\n",
    ...    old_lines = ["foo"],
    ...    new_lines = ["bar"],
    ... )
    >>> diff_lines(rfd)
    ['--- <path>', '+++ <path>', '@@ -1 +1 @@', '-foo', '+bar']
    """
    lines = difflib.unified_diff(
        a=rfd.old_lines, b=rfd.new_lines, lineterm="", fromfile=rfd.path, tofile=rfd.path
    )
    return list(lines)


def diff(new_version: str, file_patterns: config.PatternsByGlob) -> str:
    r"""Generate diffs of rewritten files.

    >>> file_patterns = {"src/pycalver/__init__.py": ['__version__ = "{pycalver}"']}
    >>> diff_str = diff("v201809.0123", file_patterns)
    >>> lines = diff_str.split("\n")
    >>> lines[:2]
    ['--- src/pycalver/__init__.py', '+++ src/pycalver/__init__.py']
    >>> assert lines[6].startswith('-__version__ = "v2')
    >>> assert not lines[6].startswith('-__version__ = "v201809.0123"')
    >>> lines[7]
    '+__version__ = "v201809.0123"'
    """

    full_diff = ""
    fh: typ.IO[str]

    for file_path, pattern_strs in sorted(_iter_file_paths(file_patterns)):
        with file_path.open(mode="rt", encoding="utf-8") as fh:
            content = fh.read()

        try:
            rfd = rfd_from_content(pattern_strs, new_version, content)
        except ValueError:
            errmsg = f"No patterns matched for '{file_path}'"
            raise ValueError(errmsg)

        rfd   = rfd._replace(path=str(file_path))
        lines = diff_lines(rfd)
        if len(lines) == 0:
            errmsg = f"No patterns matched for '{file_path}'"
            raise ValueError(errmsg)

        full_diff += "\n".join(lines) + "\n"

    full_diff = full_diff.rstrip("\n")
    return full_diff


def rewrite(new_version: str, file_patterns: config.PatternsByGlob) -> None:
    """Rewrite project files, updating each with the new version."""
    fh: typ.IO[str]

    for file_data in iter_rewritten(file_patterns, new_version):
        new_content = file_data.line_sep.join(file_data.new_lines)
        with io.open(file_data.path, mode="wt", encoding="utf-8") as fh:
            fh.write(new_content)



import copy

from pycalver import config
from pycalver import rewrite

from . import util


REWRITE_FIXTURE = """
# SPDX-License-Identifier: MIT
__version__ = "v201809.0002-beta"
"""


def test_rewrite_lines():
    old_lines = REWRITE_FIXTURE.splitlines()
    patterns  = ['__version__ = "{pycalver}"']
    new_lines = rewrite.rewrite_lines(patterns, "v201911.0003", old_lines)

    assert len(new_lines) == len(old_lines)
    assert "v201911.0003" not in "\n".join(old_lines)
    assert "v201911.0003" in "\n".join(new_lines)


def test_rewrite_final():
    # Patterns written with {release_tag} placeholder preserve
    # the release tag even if the new version is -final

    old_lines = REWRITE_FIXTURE.splitlines()
    patterns  = ['__version__ = "v{year}{month}.{build_no}-{release_tag}"']
    new_lines = rewrite.rewrite_lines(patterns, "v201911.0003", old_lines)

    assert len(new_lines) == len(old_lines)
    assert "v201911.0003" not in "\n".join(old_lines)
    assert "None" not in "\n".join(new_lines)
    assert "v201911.0003-final" in "\n".join(new_lines)


def test_iter_file_paths():
    with util.Project(project="a") as project:
        ctx = config.init_project_ctx(project.dir)
        cfg = config.parse(ctx)
        assert cfg

        file_paths = {
            str(file_path) for file_path, patterns in rewrite._iter_file_paths(cfg.file_patterns)
        }

    assert file_paths == {"pycalver.toml", "README.md"}


def test_iter_file_globs():
    with util.Project(project="b") as project:
        ctx = config.init_project_ctx(project.dir)
        cfg = config.parse(ctx)
        assert cfg

        file_paths = {
            str(file_path) for file_path, patterns in rewrite._iter_file_paths(cfg.file_patterns)
        }

    assert file_paths == {
        "setup.cfg",
        "setup.py",
        "README.rst",
        "src/module_v1/__init__.py",
        "src/module_v2/__init__.py",
    }


def test_error_bad_path():
    with util.Project(project="b") as project:
        ctx = config.init_project_ctx(project.dir)
        cfg = config.parse(ctx)
        assert cfg

        (project.dir / "setup.py").unlink()
        try:
            list(rewrite._iter_file_paths(cfg.file_patterns))
            assert False, "expected ValueError"
        except ValueError as ex:
            assert "setup.py" in str(ex)


def test_error_bad_pattern():
    with util.Project(project="b") as project:
        ctx = config.init_project_ctx(project.dir)
        cfg = config.parse(ctx)
        assert cfg

        patterns = copy.deepcopy(cfg.file_patterns)
        patterns["setup.py"] = patterns["setup.py"][0] + "invalid"

        try:
            list(rewrite.diff("v201809.1234", patterns))
            assert False, "expected ValueError"
        except ValueError as ex:
            assert "setup.py" in str(ex)
